/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "midori.h"

// -- configuration related

static CConfig* config_load(const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    CConfig* config = g_new0(CConfig, 1);

    GKeyFile* keyFile = g_key_file_new();
    g_key_file_load_from_file(keyFile, sFilename
     , G_KEY_FILE_KEEP_COMMENTS, error);
    /*g_key_file_load_from_data_dirs(keyFile, sFilename, NULL
     , G_KEY_FILE_KEEP_COMMENTS, error);*/
    #define GET_INT(var, key, default) \
     var = sokoke_key_file_get_integer_default( \
     keyFile, "browser", key, default, NULL)
    #define GET_STR(var, key, default) \
     var = sokoke_key_file_get_string_default( \
     keyFile, "browser", key, default, NULL)
    GET_INT(config->iStartup, "Startup", 1);
    GET_STR(config->sHomepage, "Homepage", "http://www.google.com");
    GET_STR(config->sLocationSearch, "LocationSearch", "http://www.google.com/search?q=%s");
    GET_INT(config->bToolbarNavigation, "ToolbarNavigation", TRUE);
    GET_INT(config->bToolbarBookmarks, "ToolbarBookmarks", FALSE);
    GET_INT(config->bToolbarStatus, "ToolbarStatus", TRUE);
    //GET_INT(config->bToolbarDownloads, "ToolbarDownloads", FALSE);
    GET_INT(config->iToolbarStyle, "ToolbarStyle", CONFIG_TOOLBAR_DEFAULT);
    GET_INT(config->bToolbarSmall, "ToolbarSmall", FALSE);
    GET_INT(config->bToolbarSearchbox, "ToolbarSearchbox", TRUE);
    GET_INT(config->bPanels, "Panels", FALSE);
    GET_INT(config->iPanelActive, "PanelActive", 0);
    GET_INT(config->iPanelPos, "PanelPos", 0);
    GET_STR(config->sPanelPageholder, "PanelPageholder", "http://www.google.com");
    GET_INT(config->iTabSize, "TabSize", 10);
    GET_INT(config->iTabClose, "TabClose", CONFIG_TABCLOSE_EVERYTAB);
    GET_INT(config->iNewPages, "NewPages", CONFIG_NEWPAGES_TAB_NEW);
    GET_INT(config->bNewTabsBackground, "NewTabsBackground", FALSE);
    GET_INT(config->bPopupsTabs, "PopupsTabs", FALSE);
    #undef GET_INT
    #undef GET_STR

    #define GET_INT(var, key, default) \
     var = sokoke_key_file_get_integer_default( \
     keyFile, "session", key, default, NULL)
    #define GET_STR(var, key, default) \
     var = sokoke_key_file_get_string_default( \
     keyFile, "session", key, default, NULL)
    
    GET_INT(config->iWinLeft  , "WinLeft", 0);
    GET_INT(config->iWinTop   , "WinTop", 0);
    GET_INT(config->iWinWidth , "WinWidth", 0);
    GET_INT(config->iWinHeight, "WinHeight", 0);
    
    #undef GET_INT
    #undef GET_STR

    config->protocols_names = g_ptr_array_new();
    g_datalist_init(&config->protocols_commands);
    gchar** protocols;
    if(protocols = g_key_file_get_keys(keyFile, "protocols", NULL, NULL))
    {
     guint i; for(i = 0; protocols[i] != NULL; i++)
     {
      gchar* sCommand = g_key_file_get_string(keyFile, "protocols"
       , protocols[i], NULL);
      g_ptr_array_add(config->protocols_names, (gpointer)protocols[i]);
      g_datalist_set_data_full(&config->protocols_commands
       , protocols[i], sCommand, g_free);
     }
     g_free(protocols);
    }

    return config;
}

gboolean config_save(CConfig* config, const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    GKeyFile* keyFile = g_key_file_new();

    g_key_file_set_integer(keyFile, "browser", "Startup", config->iStartup);
    g_key_file_set_string (keyFile, "browser", "Homepage", config->sHomepage);
    g_key_file_set_string (keyFile, "browser", "LocationSearch", config->sLocationSearch);
    g_key_file_set_integer(keyFile, "browser", "ToolbarNavigation", config->bToolbarNavigation);
    g_key_file_set_integer(keyFile, "browser", "ToolbarBookmarks", config->bToolbarBookmarks);
    //g_key_file_set_integer(keyFile, "browser", "ToolbarDownloads", config->bToolbarDownloads);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStatus", config->bToolbarStatus);
    g_key_file_set_integer(keyFile, "browser", "ToolbarStyle", config->iToolbarStyle);
    g_key_file_set_integer(keyFile, "browser", "ToolbarSmall", config->bToolbarSmall);
    g_key_file_set_integer(keyFile, "browser", "ToolbarSearchbox", config->bToolbarSearchbox);
    g_key_file_set_integer(keyFile, "browser", "Panels", config->bPanels);
    g_key_file_set_integer(keyFile, "browser", "PanelActive", config->iPanelActive);
    g_key_file_set_integer(keyFile, "browser", "PanelPos", config->iPanelPos);
    g_key_file_set_string (keyFile, "browser", "PanelPageholder", config->sPanelPageholder);
    g_key_file_set_integer(keyFile, "browser", "TabSize", config->iTabSize);
    g_key_file_set_integer(keyFile, "browser", "TabClose", config->iTabClose);
    g_key_file_set_integer(keyFile, "browser", "NewPages", config->iNewPages);
    g_key_file_set_integer(keyFile, "browser", "NewTabsBackground", config->bNewTabsBackground);
    g_key_file_set_integer(keyFile, "browser", "PopupsTabs", config->bPopupsTabs);

    //g_key_file_set_string(keyFile, "content", "LoadImages", config->bLoadImages);
    //g_key_file_set_string(keyFile, "content", "PrintBackgrounds", config->bPrintBackgrounds);
    //g_key_file_set_string(keyFile, "content", "JavaScript", config->bJavaScript);
    //g_key_file_set_string(keyFile, "content", "StandardFont", config->sStandardFont);
    //g_key_file_set_string(keyFile, "content", "SerifFont", config->sSerifFont);
    //g_key_file_set_string(keyFile, "content", "SansFont", config->sSansFontFont);
    //g_key_file_set_string(keyFile, "content", "FixedFont", config->sFixedFont);
    //g_key_file_set_string(keyFile, "content", "MinFontSize", config->iMinFontSize);

    g_key_file_set_integer(keyFile, "session", "WinLeft", config->iWinLeft);
    g_key_file_set_integer(keyFile, "session", "WinTop", config->iWinTop);
    g_key_file_set_integer(keyFile, "session", "WinWidth", config->iWinWidth);
    g_key_file_set_integer(keyFile, "session", "WinHeight", config->iWinHeight);

    guint i; for(i = 0; i < config->protocols_names->len; i++)
    {
     gchar* sProtocol = (gchar*)g_ptr_array_index(config->protocols_names, i);
     gchar* sCommand = g_datalist_get_data(&config->protocols_commands, sProtocol);
     g_key_file_set_string(keyFile, "protocols", sProtocol, sCommand);
     g_free(sProtocol);
    }
    g_datalist_clear(&config->protocols_commands);
    g_ptr_array_free(config->protocols_names, TRUE);

    gboolean bSaved = sokoke_key_file_save_to_file(keyFile, sFilename, error);
    g_key_file_free(keyFile);

    return bSaved;
}

// -- stock icons

void stock_items_init(void)
{
    // Based on ephy_stock_icons_init from lib/ephy-stock-icons.c from epiphany.
    // Copyright (C) 2002 Jorn Baayen
    static const gchar* icon_theme_items[] =
    {
       STOCK_LOCK_OPEN
     , STOCK_LOCK_SECURE
     , STOCK_LOCK_BROKEN
     , STOCK_SCRIPT
     , STOCK_THEME
     , STOCK_USER_TRASH
    };
    static const GtkStockItem items[] =
    {
       { STOCK_BOOKMARK,       "Bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARK_NEW,   "New bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARKS,      "_Bookmarks", 0, 0, NULL }
     , { STOCK_DOWNLOADS,      "_Downloads", 0, 0, NULL }
     , { STOCK_CONSOLE,        "_Console", 0, 0, NULL }
     , { STOCK_EXTENSIONS,     "_Extensions", 0, 0, NULL }
     , { STOCK_FORM_FILL,      "_Form fill", 0, 0, NULL }
     , { STOCK_HISTORY,        "History", 0, 0, NULL }
     , { STOCK_HOMEPAGE,       "Homepage", 0, 0, NULL }
     , { STOCK_LOCATION,       "Location entry", 0, 0, NULL }
     , { STOCK_NEWSFEED,       "Newsfeed", 0, 0, NULL }
     , { STOCK_PLUGINS,        "_Plugins", 0, 0, NULL }
     , { STOCK_POPUPS_BLOCKED, "Blocked popups", 0, 0, NULL }
     , { STOCK_SOURCE_VIEW,    "View source", 0, 0, NULL }
     , { STOCK_TAB_CLOSE,      "Close tab", 0, 0, NULL }
     , { STOCK_TAB_NEW,        "New tab", 0, 0, NULL }
     , { STOCK_WINDOW_CLOSE,   "Close window", 0, 0, NULL }
     , { STOCK_WINDOW_NEW,     "New window", 0, 0, NULL }
     #if !GTK_CHECK_VERSION(2, 10, 0)
     , { GTK_STOCK_SELECT_ALL, "Select _All", 0, 0, NULL }
     #endif
    };
    // TODO: Use fallback icons with _ suffix
    GtkIconFactory* factory = gtk_icon_factory_new ();
    guint i;
    for(i = 0; i < (guint)G_N_ELEMENTS(items); i++)
    {
     GtkIconSource* icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name(icon_source, items[i].stock_id);
     GtkIconSet* icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, items[i].stock_id, icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_stock_add_static(items, G_N_ELEMENTS(items));

    for(i = 0; i < (guint)G_N_ELEMENTS(icon_theme_items); i++)
    {
     GtkIconSource* icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name(icon_source, icon_theme_items[i]);
     GtkIconSet* icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, icon_theme_items[i], icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_icon_factory_add_default(factory);
    g_object_unref(factory);
}

// -- main function

int main(int argc, char* *argv)
{
    DEBUG_FUNC

    CRASHSAFE_INIT
    g_set_application_name(PACKAGE_NAME);

    // Parse cli options
    gint repeats = 2;
    gboolean bVersion = FALSE;
    GOptionEntry entries[] =
    {
     { "repeats", 'r', 0, G_OPTION_ARG_INT, &repeats, "An unused value", "N" },
     { "version", 'v', 0, G_OPTION_ARG_NONE, &bVersion, "Display program version", NULL }
    };

    GError* error = NULL;
    if(!gtk_init_with_args(&argc, &argv, "[URI]", entries
     , NULL/*GETTEXT_PACKAGE*/, &error))
    {
     g_error_free(error);
     return 1;
    }

    if(bVersion)
    {
     g_print(PACKAGE_STRING " - Copyright (c) 2007 Christian Dywan\n\n"
             "GTK+2:      " GTK_VER "\n"
	     "WebKit:     " WEBKIT_VER "\n"
	     "Libsexy:    " LIBSEXY_VER "\n"
             "GetText:    N/A\n"
	     "\n"
	     "CrashSafe:  " SOKOKE_CRASHSAFE_ "\n"
	     "Debugging:  " SOKOKE_DEBUG_ "\n"
	     "Engine:     " MIDORI_ENGINE_ "\n"
             "\n"
             "Please report comments, suggestions and bugs to:\n"
             "\t" PACKAGE_BUGREPORT "\n"
             "Check for new versions at:\n"
             "\thttp://software.twotoasts.de\n");
     return 0;
    }

    // Load configuration files
    // TODO: What about default config in a global config folder?
    gchar* sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    gchar* sConfigFile = g_build_filename(sConfig, "config", NULL);
    config = config_load(sConfigFile, &error); g_free(sConfigFile);
    if(error != NULL)
    {
     g_print("Notice: No configuration loaded. %s", error->message);
     g_error_free(error);
    }
    gchar* sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_load(sAccelsFile); g_free(sAccelsFile);
    /*gchar* sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    bookmarks = g_bookmark_file_new();
    if(!g_bookmark_file_load_from_file(bookmarks, sBookmarkFile, &error))
     g_warning("Bookmarks couldn't be loaded. %s", error->message); g_free(sBookmarkFile);*/
    g_free(sConfig);

    //browser_list = g_list_alloc();
    accel_group = gtk_accel_group_new();

    stock_items_init();
    #ifdef MIDORI_WEBKIT
    webkit_init();
    #endif
    browser_list = NULL;
    browser_count = GUINT_TO_POINTER(0);

    // TODO: Handle any number of separate uris from argv
    gchar* sHomepage = argc > 1 ? argv[1]
     : (config->iStartup == 1 ? config->sHomepage : "http://about:blank");
    // Open as many tabs as we have uris, seperated by pipes
    CBrowser* browser = NULL;
    gchar* sUri = strtok(g_strdup(sHomepage), "|");
    while(sUri != NULL)
    {
     browser = browser_new(browser);
     /* TODO: Use g_idle_add to prevent code from being called
     before the main loop is running */
     document_load_uri(browser->document, magic_uri(sUri, config->sLocationSearch));
     sUri = strtok(NULL, "|");
    } g_free(sUri);

    gtk_main();
    DEBUG_FUNC

    // Save configuration files
    sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    sConfigFile = g_build_filename(sConfig, "config", NULL);
    if(!config_save(config, sConfigFile, &error))
    {
     g_warning("Configuration couldn't be saved. %s", error->message);
     g_error_free(error);
    }
    g_free(sConfigFile);
    sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_save(sAccelsFile); g_free(sAccelsFile);
    /*sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    if(!g_bookmark_file_to_file(bookmarks, sBookmarkFile, NULL))
     g_warning("Bookmarks couldn't be saved."); g_free(sBookmarkFile);*/
    g_free(sConfig);
    return 0;
}
