/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "prefs.h"

gboolean on_preferences_homepage_focus_out(GtkWidget* widget, GdkEventFocus event
 , CBrowser* browser)
{
    g_free(config->sHomepage);
    config->sHomepage = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    return FALSE;
}

void on_preferences_loadonstartup_changed(GtkWidget* widget, CBrowser* browser)
{
    config->iStartup = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
}

void on_preferences_newpages_changed(GtkWidget* widget, CBrowser* browser)
{
    config->iNewPages = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
}

void on_preferences_tabsbackground_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->bNewTabsBackground = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}

void on_preferences_popupstabs_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->bPopupsTabs = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}

void on_preferences_toolbarstyle_changed(GtkWidget* widget, CBrowser* browser)
{
    config->iToolbarStyle = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbarstyle(config->iToolbarStyle));
}

void on_preferences_toolbarsmall_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->bToolbarSmall = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbariconsize(config->bToolbarSmall));
}

void on_preferences_closebuttonsontabs_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->iTabClose = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    GList* browsers = browser_list;
    CBrowser* curBrowser;
    do
    {
     curBrowser = (CBrowser*)browsers->data;
     sokoke_widget_set_visible(curBrowser->document_close, config->iTabClose);
    }
    while(browsers = g_list_next(browsers)); g_list_free(browsers);
}

void on_preferences_toolbarsearchbox_toggled(GtkWidget* widget, CBrowser* browser)
{
    config->bToolbarSearchbox = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    sokoke_widget_set_visible(browser->navibar_search, config->bToolbarSearchbox);
}

gboolean on_preferences_locationsearch_focus_out(GtkWidget* widget, GdkEventFocus event
 , CBrowser* browser)
{
    g_free(config->sLocationSearch);
    config->sLocationSearch = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    return FALSE;
}

void on_preferences_protocols_render_icon(GtkTreeViewColumn* column
 , GtkCellRenderer* renderer, GtkTreeModel* model, GtkTreeIter* iter
 , GtkWidget* treeview)
{
    DEBUG_FUNC
    gchar* sCommand;
    gtk_tree_model_get(model, iter, PROTOCOLS_COL_COMMAND, &sCommand, -1);

    // TODO: Would it be better, to not do this on every redraw?
    // Determine the actual binary to be able to display an icon
    gchar* sBinary = NULL;
    if(sCommand) sBinary = strtok(sCommand, " ");
    if(sBinary)
    {
     gchar* sPath; if(sPath = g_find_program_in_path(sBinary))
     {
      GtkIconTheme* icon_theme = get_icon_theme(treeview);
      if(g_path_is_absolute(sBinary))
      {
       g_free(sPath); sPath = g_path_get_basename(sBinary); sBinary = sPath;
      }
      // TODO: Is it good to just display nothing if there is no icon?
      if(!gtk_icon_theme_has_icon(icon_theme, sBinary)) sBinary = NULL;
      #if GTK_CHECK_VERSION(2, 10, 0)
      g_object_set(renderer, "icon-name", sBinary, NULL);
      #else
      // TODO: Assign a pixbuf manually, icon-name is not available < 2.10
      g_object_set(renderer, "stock-id", GTK_STOCK_FILE, NULL);
      #endif
      g_free(sPath);
     }
     else
     {
      g_object_set(renderer, "icon-name", NULL, NULL);//...
      g_object_set(renderer, "stock-id", GTK_STOCK_DIALOG_ERROR, NULL);
     }
    }
    else
    {
     // We need to reset the icon
     #if GTK_CHECK_VERSION(2, 10, 0)
     g_object_set(renderer, "icon-name", NULL, NULL);
     #else
     g_object_set(renderer, "stock-id", NULL, NULL);
     #endif
    }
    g_free(sCommand);
}

static void on_preferences_protocols_edited(GtkCellRendererText* renderer
 , gchar* sPath, gchar* sTextNew, GtkWidget* treeview)
{
    DEBUG_FUNC
    GtkTreeModel* model = gtk_tree_view_get_model(GTK_TREE_VIEW(treeview));
    GtkTreeIter iter;
    gtk_tree_model_get_iter_from_string(model, &iter, sPath);
    gtk_list_store_set(GTK_LIST_STORE(model), &iter
     , PROTOCOLS_COL_COMMAND, sTextNew, -1);
    gchar* sProtocol;
    gtk_tree_model_get(model, &iter, PROTOCOLS_COL_NAME, &sProtocol, -1);
    g_datalist_set_data_full(&config->protocols_commands
     , sProtocol, g_strdup(sTextNew), g_free);
}

static void on_preferences_protocols_add_clicked(GtkWidget* widget
 , CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: Don't use globals here
    gchar* sProtocol = gtk_combo_box_get_active_text(GTK_COMBO_BOX(combo_protocol));
    GtkTreeModel* liststore = gtk_tree_view_get_model(GTK_TREE_VIEW(treeview_protocols));
    gtk_list_store_insert_with_values(GTK_LIST_STORE(liststore), NULL, G_MAXINT
        , PROTOCOLS_COL_NAME, sProtocol
        , PROTOCOLS_COL_COMMAND, "", -1);
    g_ptr_array_add(config->protocols_names, (gpointer)sProtocol);
    g_datalist_set_data_full(&config->protocols_commands
     , sProtocol, g_strdup(""), g_free);
}

void on_action_preferences_activate(GtkAction* action, CBrowser* browser)
{
    DEBUG_FUNC
    // Show the preferences dialog. Create it if necessary.
    static GtkWidget* dialog;
    if(GTK_IS_DIALOG(dialog))
     gtk_window_present(GTK_WINDOW(dialog));
    else
    {
     gchar* sDialogTitle = g_strdup_printf("%s Preferences", PACKAGE_NAME);
     dialog = gtk_dialog_new_with_buttons(sDialogTitle
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 , GTK_STOCK_HELP
                                 , GTK_RESPONSE_HELP
                                 , GTK_STOCK_CLOSE
                                 , GTK_RESPONSE_CLOSE
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(dialog), GTK_STOCK_PREFERENCES);
     gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
     // TODO: Implement some kind of help function
     gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog)
      , GTK_RESPONSE_HELP, FALSE); //...
     g_signal_connect(GTK_OBJECT(dialog), "response"
      , GTK_SIGNAL_FUNC(gtk_widget_destroy), dialog);
     // TODO: Do we want tooltips for explainations or can we omit that?
     // TODO: We need mnemonics
     // TODO: Take multiple windows into account when applying changes
     GtkWidget* xfce_heading;
     // TODO: "Assignment creates pointer without type conversion"
     if(xfce_heading = sokoke_xfce_header_new(
      gtk_window_get_icon_name(GTK_WINDOW(browser->window)), sDialogTitle))
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox)
      , xfce_heading, FALSE, FALSE, 0);
     g_free(sDialogTitle);
     GtkWidget* notebook = gtk_notebook_new();
     gtk_container_set_border_width(GTK_CONTAINER(notebook), 6);
     GtkSizeGroup* sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
     GtkSizeGroup* sizegroup2 = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
     GtkWidget* page; GtkWidget* frame; GtkWidget* table; GtkWidget* align;
     GtkWidget* button; GtkWidget* checkbutton; GtkWidget* colorbutton;
     GtkWidget* combobox; GtkWidget* entry; GtkWidget* hbox; GtkWidget* spinbutton;
     #define PAGE_NEW(__label) page = gtk_vbox_new(FALSE, 0);\
      gtk_container_set_border_width(GTK_CONTAINER(page), 5);\
      gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page, gtk_label_new(__label))
     #define FRAME_NEW(__label) frame = sokoke_hig_frame_new(__label);\
      gtk_container_set_border_width(GTK_CONTAINER(frame), 5);\
      gtk_box_pack_start(GTK_BOX(page), frame, FALSE, FALSE, 0);
     #define TABLE_NEW(__rows, __cols) table = gtk_table_new(__rows, __cols, FALSE);\
      gtk_container_set_border_width(GTK_CONTAINER(table), 5);\
      gtk_container_add(GTK_CONTAINER(frame), table);
     #define WIDGET_ADD(__widget, __left, __right, __top, __bottom)\
      gtk_table_attach(GTK_TABLE(table), __widget\
       , __left, __right, __top, __bottom\
       , 0, GTK_FILL, 8, 2)
     #define FILLED_ADD(__widget, __left, __right, __top, __bottom)\
      gtk_table_attach(GTK_TABLE(table), __widget\
       , __left, __right, __top, __bottom\
       , GTK_EXPAND | GTK_FILL, GTK_FILL, 8, 2)
     #define INDENTED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      gtk_size_group_add_widget(sizegroup, align);\
      WIDGET_ADD(align, __left, __right, __top, __bottom)
     #define SEMI_INDENTED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      gtk_size_group_add_widget(sizegroup2, align);\
      WIDGET_ADD(align, __left, __right, __top, __bottom)
     #define SPANNED_ADD(__widget, __left, __right, __top, __bottom)\
      align = gtk_alignment_new(0, 0.5, 0, 0);\
      gtk_container_add(GTK_CONTAINER(align), __widget);\
      FILLED_ADD(align, __left, __right, __top, __bottom)
     // Page "General"
     PAGE_NEW("General");
     FRAME_NEW("Startup");
     TABLE_NEW(2, 2);
     INDENTED_ADD(gtk_label_new("Load on startup"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Blank page", "Homepage"/*, "Last open pages"*/, NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), config->iStartup);
     g_signal_connect(combobox, "changed"
     , G_CALLBACK(on_preferences_loadonstartup_changed), browser);
     FILLED_ADD(combobox, 1, 2, 0, 1);
     INDENTED_ADD(gtk_label_new("Homepage"), 0, 1, 1, 2);
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), config->sHomepage);
     g_signal_connect(entry, "focus-out-event"
     , G_CALLBACK(on_preferences_homepage_focus_out), browser);
     FILLED_ADD(entry, 1, 2, 1, 2);
     // TODO: We need something like "use current website"
     FRAME_NEW("Downloads");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Download folder"), 0, 1, 0, 1);
     GtkWidget* filebutton = gtk_file_chooser_button_new(
      "Choose download folder", GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
     gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(filebutton)
      , g_get_home_dir()); //...
     gtk_widget_set_sensitive(filebutton, FALSE); //...
     FILLED_ADD(filebutton, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show a notification window for finished downloads");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     FRAME_NEW("Languages");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Preferred languages"), 0, 1, 0, 1);
     entry = gtk_entry_new();
     // TODO: Make sth like get_browser_languages_default filtering encodings and C out
     // TODO: Provide a real ui with real language names (iso-codes)
     const gchar* const* sLanguages = g_get_language_names();
     //for (i = 0; sLanguages[i] != NULL; i++) g_strdup(sLanguages[i]);
     //for (str = strings; *str; str++);
     gchar* sLanguagesPreferred = g_strjoinv(",", (gchar**)sLanguages);
     gtk_entry_set_text(GTK_ENTRY(entry), sLanguagesPreferred/*config->sLanguagesPreferred*/);
     g_free(sLanguagesPreferred);
     gtk_widget_set_sensitive(entry, FALSE); //...
     FILLED_ADD(entry, 1, 2, 0, 1);

     // Page "Appearance"
     PAGE_NEW("Appearance");
     FRAME_NEW("Font settings");
     TABLE_NEW(5, 2);
     INDENTED_ADD(gtk_label_new("Standard font"), 0, 1, 0, 1);
     button = gtk_font_button_new_with_font("Sans 10"/*config->sFontStandard*/);
     gtk_widget_set_sensitive(button, FALSE); //...
     FILLED_ADD(button, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Use theme _default font");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Minimum font size"), 0, 1, 2, 3);
     spinbutton = gtk_spin_button_new_with_range(6, 30, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 8/*config->iFontSizeMin*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     SPANNED_ADD(spinbutton, 1, 2, 2, 3);
     INDENTED_ADD(gtk_label_new("Default encoding"), 0, 1, 3, 4);
     combobox = gtk_combo_box_new_text();
     const gchar* encoding = NULL; g_get_charset(&encoding);
     // TODO: Fallback to utf-8 if the encoding is not sane (e.g. when lang=C)
     gchar* sEncodingDefault = g_strdup_printf("Default (%s)", encoding);
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , sEncodingDefault, "Chinese", "Greek", "Japanese (SHIFT_JIS)"
      , "Korean", "Russian", "Unicode (UTF-8)", "Western (ISO-8859-1)", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 3, 4);
     button = gtk_button_new_with_label("Advanced settings");
     gtk_widget_set_sensitive(button, FALSE); //...
     WIDGET_ADD(button, 1, 2, 3, 4);
     FRAME_NEW("Default colors");
     TABLE_NEW(2, 4);
     SEMI_INDENTED_ADD(gtk_label_new("Text color"), 0, 1, 0, 1);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 1, 2, 0, 1);
     SEMI_INDENTED_ADD(gtk_label_new("Background color"), 2, 3, 0, 1);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 3, 4, 0, 1);
     SEMI_INDENTED_ADD(gtk_label_new("Normal link color"), 0, 1, 1, 2);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 1, 2, 1, 2);
     SEMI_INDENTED_ADD(gtk_label_new("Visited link color"), 2, 3, 1, 2);
     colorbutton = gtk_color_button_new();
     gtk_widget_set_sensitive(colorbutton, FALSE); //...
     WIDGET_ADD(colorbutton, 3, 4, 1, 2);

     // Behavior
     PAGE_NEW("Behavior");
     FRAME_NEW("Browsing");
     TABLE_NEW(3, 2);
     INDENTED_ADD(gtk_label_new_with_mnemonic("Open _new pages in"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "New tab", "New window", "Current tab", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), config->iNewPages);
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic("Open new tabs in the _background");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton), config->bNewTabsBackground);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_preferences_tabsbackground_toggled), browser);
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("Open _popups in tabs");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton), config->bPopupsTabs);
     g_signal_connect(checkbutton, "toggled"
      , G_CALLBACK(on_preferences_popupstabs_toggled), browser);
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 2, 3);

     // Page "Interface"
     PAGE_NEW("Interface");
     FRAME_NEW("Navigationbar");
     TABLE_NEW(3, 2);
     INDENTED_ADD(gtk_label_new_with_mnemonic("_Toolbar style"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Default", "Icons", "Text", "Both", "Both horizontal", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), config->iToolbarStyle);
     g_signal_connect(combobox, "changed"
     , G_CALLBACK(on_preferences_toolbarstyle_changed), browser);
     FILLED_ADD(combobox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic("Show small _icons");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton), config->bToolbarSmall);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_preferences_toolbarsmall_toggled), browser);
     SPANNED_ADD(checkbutton, 0, 1, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("Show _searchbox");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton), config->bToolbarSearchbox);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_preferences_toolbarsearchbox_toggled), browser);
     SPANNED_ADD(checkbutton, 1, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("Show _new tab button");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 1, 2, 3);
     checkbutton = gtk_check_button_new_with_mnemonic("Show _closed tabs button");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 1, 2, 2, 3);
     FRAME_NEW("Miscellaneous");
     TABLE_NEW(3, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Show close _buttons on tabs");
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton)
      , config->iTabClose);
     g_signal_connect(checkbutton, "toggled"
     , G_CALLBACK(on_preferences_closebuttonsontabs_toggled), browser);
     SPANNED_ADD(checkbutton, 0, 2, 0, 1);
     INDENTED_ADD(gtk_label_new_with_mnemonic("Tabbar _placement"), 0, 1, 1, 2);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "Left", "Top", "Right", "Bottom", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 1); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 1, 2);
     INDENTED_ADD(gtk_label_new_with_mnemonic("_Location search engine"), 0, 1, 2, 3);
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), config->sLocationSearch);
     g_signal_connect(entry, "focus-out-event"
     , G_CALLBACK(on_preferences_locationsearch_focus_out), browser);
     FILLED_ADD(entry, 1, 2, 2, 3);

     // Page "Network"
     PAGE_NEW("Network");
     FRAME_NEW("Proxy Server");
     gtk_widget_set_sensitive(frame, FALSE); //...
     TABLE_NEW(5, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("_Custom proxy server");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 0, 1);
     hbox = gtk_hbox_new(FALSE, 4);
     INDENTED_ADD(gtk_label_new_with_mnemonic("_Host/ Port"), 0, 1, 1, 2);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_entry_new(), FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_spin_button_new_with_range(0, 65535, 1), FALSE, FALSE, 0);
     FILLED_ADD(hbox, 1, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Proxy requires authentication");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     // TODO The proxy user and pass need to be indented further
     SPANNED_ADD(checkbutton, 0, 2, 2, 3);
     INDENTED_ADD(gtk_label_new("Username"), 0, 1, 3, 4);
     FILLED_ADD(gtk_entry_new(), 1, 2, 3, 4);
     INDENTED_ADD(gtk_label_new("Password"), 0, 1, 4, 5);
     FILLED_ADD(gtk_entry_new(), 1, 2, 4, 5);
     FRAME_NEW("Cache");
     TABLE_NEW(1, 2);
     INDENTED_ADD(gtk_label_new("Cache size"), 0, 1, 0, 1);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 10000, 10);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 100/*config->iCacheSize*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("MB"), FALSE, FALSE, 0);
     button = gtk_button_new_with_label("Clear cache");
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 4);
     SPANNED_ADD(hbox, 1, 2, 0, 1);

     // Page "Privacy"
     PAGE_NEW("Privacy");
     FRAME_NEW("Cookies");
     TABLE_NEW(3, 2);
     INDENTED_ADD(gtk_label_new("Accept cookies"), 0, 1, 0, 1);
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "All cookies", "Session cookies", "None", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0); //...
     gtk_widget_set_sensitive(combobox, FALSE); //...
     FILLED_ADD(combobox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Allow cookies from the original website only");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     INDENTED_ADD(gtk_label_new("Maximum cookie age"), 0, 1, 2, 3);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 360, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 30/*config->iCookieAgeMax*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("days"), FALSE, FALSE, 0);
     button = gtk_button_new_with_label("View cookies");
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 4);
     SPANNED_ADD(hbox, 1, 2, 2, 3);
     FRAME_NEW("History");
     TABLE_NEW(3, 2);
     checkbutton = gtk_check_button_new_with_mnemonic("Remember my visited pages");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 1, 0, 1);
     hbox = gtk_hbox_new(FALSE, 4);
     spinbutton = gtk_spin_button_new_with_range(0, 360, 1);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton), 30/*config->iHistoryAgeMax*/);
     gtk_widget_set_sensitive(spinbutton, FALSE); //...
     gtk_box_pack_start(GTK_BOX(hbox), spinbutton, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("days"), FALSE, FALSE, 0);
     SPANNED_ADD(hbox, 1, 2, 0, 1);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Remember my form inputs");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 1, 2);
     checkbutton = gtk_check_button_new_with_mnemonic
      ("Remember my downloaded files");
     gtk_widget_set_sensitive(checkbutton, FALSE); //...
     SPANNED_ADD(checkbutton, 0, 2, 2, 3);

     // Page "Programs"
     PAGE_NEW("Programs");
     FRAME_NEW("External programs");
     TABLE_NEW(3, 2);
     GtkTreeView* treeview; GtkTreeViewColumn* column;
     GtkCellRenderer* renderer_text; GtkCellRenderer* renderer_pixbuf;
     GtkListStore* liststore = gtk_list_store_new(PROTOCOLS_COL_N
      , G_TYPE_STRING, G_TYPE_STRING);
     treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(liststore));
     renderer_text = gtk_cell_renderer_text_new();
     renderer_pixbuf = gtk_cell_renderer_pixbuf_new();
     column = gtk_tree_view_column_new_with_attributes(
      "Protocol", renderer_text, "text", PROTOCOLS_COL_NAME, NULL);
     gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
     column = gtk_tree_view_column_new();
     gtk_tree_view_column_set_title(column, "Command");
     gtk_tree_view_column_pack_start(column, renderer_pixbuf, FALSE);
     gtk_tree_view_column_set_cell_data_func(column, renderer_pixbuf
     , (GtkTreeCellDataFunc)on_preferences_protocols_render_icon, treeview, NULL);
     renderer_text = gtk_cell_renderer_text_new();
     g_object_set(G_OBJECT(renderer_text), "editable", TRUE, NULL);
     g_signal_connect(GTK_OBJECT(renderer_text), "edited"
      ,G_CALLBACK(on_preferences_protocols_edited), treeview);
     gtk_tree_view_column_pack_start(column, renderer_text, TRUE);
     gtk_tree_view_column_add_attribute(column, renderer_text, "text", PROTOCOLS_COL_COMMAND);
     gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
     GtkWidget* scrolled = gtk_scrolled_window_new(NULL, NULL);
     gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
     , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
     gtk_container_add(GTK_CONTAINER(scrolled), treeview);
     FILLED_ADD(sokoke_frame_new_child_shadow(scrolled, GTK_SHADOW_ETCHED_IN), 0, 2, 0, 2);
     guint i; for(i = 0; i < config->protocols_names->len; i++)
     {
      gchar* sProtocol = (gchar*)g_ptr_array_index(config->protocols_names, i);
      // TODO: We might want to determine 'default' programs somehow
      // TODO: Any way to make it easier to add eg. only a mail client? O_o
      const gchar* sCommand = g_datalist_get_data(&config->protocols_commands, sProtocol);
      gtk_list_store_insert_with_values(GTK_LIST_STORE(liststore), NULL, i
       , PROTOCOLS_COL_NAME, sProtocol
       , PROTOCOLS_COL_COMMAND, sCommand, -1);
     }
     g_object_unref(liststore);
     // TODO: There should be a "remove" button
     // TODO: Disable on change if we have the protocol
     combobox = gtk_combo_box_new_text();
     sokoke_combo_box_add_strings(GTK_COMBO_BOX(combobox)
      , "download", "e2dk", "feed", "ftp", "irc", "mailto"
      , "news", "tel", "torrent", "view-source", NULL);
     gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 1);
     SPANNED_ADD(combobox, 0, 1, 2, 3);
     button = gtk_button_new_from_stock(GTK_STOCK_ADD);
     // TODO: Don't use the two globals
     treeview_protocols = treeview; combo_protocol = combobox;
     g_signal_connect(GTK_OBJECT(button), "clicked"
      ,G_CALLBACK(on_preferences_protocols_add_clicked), browser);
     SPANNED_ADD(button, 1, 2, 2, 3);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox)
      , notebook, FALSE, FALSE, 0);
     gtk_widget_show_all(dialog);
    }
}
