/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#ifndef __UI_H__
#define __UI_H__ 1

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

typedef struct _CConfig
{
    // TODO: Would it be cleaner/ reasonable to use a GData* for all this?
    guint iStartup; // 0: blank, 1: homepage, 2: last open pages
    gchar* sHomepage;
    gchar* sLocationSearch;
    gboolean bToolbarNavigation;
    gboolean bToolbarBookmarks;
    //gboolean bToolbarDownloads;
    gboolean bToolbarStatus;
    guint iToolbarStyle;
    guint bToolbarSmall;
    gboolean bPanels;
    guint iPanelActive;
    guint iPanelPos;
    gchar* sPanelPageholder;
    // TODO: What about this? Support it or fix tab shrinking and drop it?
    guint iTabSize; // tab size in charcters
    guint iTabClose;
    gboolean bToolbarSearchbox;
    guint iNewPages; // where to open new pages
    gboolean bNewTabsBackground;
    gboolean bPopupsTabs; // open popups in tabs?

    gint iWinLeft;
    gint iWinTop;
    gint iWinWidth;
    gint iWinHeight;

    GPtrArray* protocols_names;
    GData*     protocols_commands;
} CConfig;

enum {
 CONFIG_TOOLBAR_DEFAULT,
 CONFIG_TOOLBAR_ICONS,
 CONFIG_TOOLBAR_TEXT,
 CONFIG_TOOLBAR_BOTH,
 CONFIG_TOOLBAR_BOTH_HORIZ
 };

enum {
 CONFIG_TABCLOSE_NONE,
 CONFIG_TABCLOSE_EVERYTAB,
 // TODO: Do we want a toolbar close button?
 //CONFIG_NEWPAGES_TOOLBAR
 };

enum {
 CONFIG_NEWPAGES_TAB_NEW,
 CONFIG_NEWPAGES_WINDOW_NEW,
 CONFIG_NEWPAGES_TAB_CURRENT
 };


// Use these until browsers are guaranteed to be created inside the main loop
#define ZOMBIE(rv) if(!GTK_IS_WIDGET(browser->menubar)) \
                    { /*g_print("* ZOMBIE %s\n", __func__);*/ return rv; }
#define ZOMBIE_DOC(rv) if(!GTK_IS_WIDGET(document)) return rv; \
                    { /*g_print("* ZOMBIE %s\n", __func__);*/ return rv; }

typedef struct _CBrowser
{
    CConfig* config;

    // shared widgets
    GtkWidget* window;
    GtkActionGroup* actiongroup;
    // menus
    GtkWidget* menubar;
    GtkWidget* menu_tabsclosed;
    GtkWidget* menu_window;
    GtkWidget* popup_document;
    GtkWidget* popup_back;
    GtkWidget* popup_forward;
    GtkWidget* throbber;
    // navibar
    GtkWidget* navibar;
    GtkWidget* icon_page;
    GtkWidget* navibar_location;
    GtkWidget* icon_search;
    GtkWidget* navibar_search;
    GtkWidget* icon_search_add;
    GtkWidget* bookmarkbar;
    // panels
    GtkWidget* panels;
    GtkWidget* panels_notebook;
    GtkWidget* panel_pageholder;
    GtkWidget* documents;
    // findbox
    GtkWidget* findbox;
    GtkWidget* findbox_text;
    GtkToolItem* findbox_case;
    GtkWidget* statusbar;
    GtkWidget* icon_link_script;
    GtkWidget* icon_link_window;
    GtkWidget* icon_network;
    GtkWidget* progress;
    GtkWidget* icon_security;
    GtkWidget* icon_newsfeed;

    // document specific widgets
    GtkWidget* document_menu;
    GtkWidget* document_icon;
    GtkWidget* document_name;
    GtkWidget* document_close;
    GtkWidget* document;

    // document specific values
    gboolean bMenubar;
    gboolean bToolbar;
    gboolean bLocation;
    gboolean bStatusbar;
    gboolean bResizable;
    gchar* sElementUri; // the element the mouse is hovering on
    gint iLoadedPercent; // -1 means "not loading"
    gint iLoadedBytes;
    gint iLoadedBytesMax;
    //UNDEFINED favicon;
    guint iSecurity;
    gchar* sStatusMessage; // message from a document
} CBrowser;

// -- globals

GList* browser_list; // Items of type 'CBrowser'
void* browser_count; // guint*
CConfig* config;
GtkAccelGroup* accel_group;
GBookmarkFile* bookmarks;

// Custom stock items

// We should distribute these
// Names should match with epiphany and/ or xdg spec
/* NOTE: Those uncommented were replaced with remotely related icons
         in order to suppress annoying warnings :D */

#define STOCK_BOOKMARK           "stock_bookmark" // "bookmark-web"
#define STOCK_BOOKMARKS          "bookmark-view"
#define STOCK_DOWNLOADS          "package" // "download"
#define STOCK_CONSOLE            "terminal" // "console" // MISSING
#define STOCK_EXTENSIONS         "extension" // MISSING
#define STOCK_FORM_FILL          "insert-text" // "form-fill" // MISSING
#define STOCK_HISTORY            "document-open-recent"
#define STOCK_HISTORY_           "history-view"
#define STOCK_LOCATION           "location-entry"
#define STOCK_NEWSFEED           "gtk-index" // "newsfeed" // MISSING
#define STOCK_PLUGINS            "plugin" // MISSING
#define STOCK_POPUPS_BLOCKED     "popup-hidden"
#define STOCK_SOURCE_VIEW        "stock_view-html-source" // MISSING
#define STOCK_TAB_CLOSE          "tab-close" // MISSING
#define STOCK_WINDOW_CLOSE       "window-close" // MISSING

// We can safely use standard icons
// Assuming that we have reliable fallback icons

#define STOCK_BOOKMARK_NEW       "bookmark-new"
#define STOCK_BOOKMARK_NEW_      "stock_add-bookmark"
#define STOCK_HOMEPAGE           GTK_STOCK_HOME
#define STOCK_IMAGE              "image-x-generic"
#define STOCK_IMAGE_             "gnome-mime-image"
#define STOCK_LOCK_OPEN          "stock_lock-open"
#define STOCK_LOCK_SECURE        "stock_lock"
#define STOCK_LOCK_BROKEN        "stock_lock-broken"
#define STOCK_NETWORK_OFFLINE    "connect_no"
#define STOCK_NETWORK_OFFLINE_   "network-offline"
#define STOCK_SCRIPT             "stock_script"
#define STOCK_SEND               "mail-send"
#define STOCK_SEND_              "stock_mail-send"
#define STOCK_TAB_NEW            "tab-new"
#define STOCK_TAB_NEW_           "stock_new-tab"
#define STOCK_THEME              "gnome-settings-theme"
#define STOCK_USER_TRASH         "user-trash"
#define STOCK_USER_TRASH_        "gnome-stock-trash"
#define STOCK_WINDOW_NEW         "window-new"
#define STOCK_WINDOW_NEW_        "stock_new-window"

// For backwards compatibility

#if !GTK_CHECK_VERSION(2, 10, 0)
#define GTK_STOCK_SELECT_ALL     "gtk-select-all"
#endif

// -- Credits

// TODO: Retrieve credits from some kind of normal text files
static const gchar* credits_authors[] = { "Christian Dywan <christian@twotoasts.de>", NULL };
static const gchar* credits_documenters/*[]*/ = /*{ */NULL/* }*/;
static const gchar* credits_artists[] = { "Nancy Runge <nancy@twotoasts.de>", NULL };

// -- Licenses

static const gchar* license_gpl_2 = \
 ("This program is free software; you can redistribute it and/or modify "
  "it under the terms of the GNU General Public License as published by "
  "the Free Software Foundation; either version 2 of the License, or "
  "(at your option) any later version."
  "\n\n"
  "This program is distributed in the hope that it will be useful, "
  "but WITHOUT ANY WARRANTY; without even the implied warranty of "
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
  "GNU General Public License for more details."
  "\n\n"
  "You should have received a copy of the GNU General Public License "
  "along with this program; if not, write to the Free Software Foundation, Inc., "
  "51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.");

// -- User interface description

static const gchar* ui_markup =
 "<ui>"
  "<menubar>"
   "<menu action='File'>"
    "<menuitem action='WindowNew'/>"
    "<menuitem action='TabNew'/>"
    "<separator/>"
    "<menuitem action='Open'/>"
    "<separator/>"
    "<menuitem action='SaveAs'/>"
    "<separator/>"
    "<menuitem action='TabClose'/>"
    "<menuitem action='WindowClose'/>"
    "<separator/>"
    "<menuitem action='PageSetup'/>"
    "<menuitem action='PrintPreview'/>"
    "<menuitem action='Print'/>"
    "<separator/>"
    "<menuitem action='PrivateBrowsing'/>"
    "<menuitem action='WorkOffline'/>"
    "<separator/>"
    "<menuitem action='Quit'/>"
   "</menu>"
   "<menu action='Edit'>"
    "<menuitem action='Undo'/>"
    "<menuitem action='Redo'/>"
    "<separator/>"
    "<menuitem action='Cut'/>"
    "<menuitem action='Copy'/>"
    "<menuitem action='Paste'/>"
    "<menuitem action='Delete'/>"
    "<menuitem action='SelectAll'/>"
    "<separator/>"
    "<menuitem action='FormFill'/>"
    "<separator/>"
    "<menuitem action='Find'/>"
    "<menuitem action='FindNext'/>"
    "<menuitem action='FindPrevious'/>"
    "<menuitem action='FindQuick'/>"
    "<separator/>"
    "<menuitem action='Preferences'/>"
   "</menu>"
   "<menu action='View'>"
    "<menuitem action='ToolbarNavigation'/>"
    "<menuitem action='ToolbarBookmarks'/>"
    // TODO: Does this belong here?
    "<menuitem action='Panels'/>"
    "<menuitem action='ToolbarDownloads'/>"
    "<menuitem action='ToolbarStatus'/>"
    "<separator/>"
    "<menuitem action='Refresh'/>"
    "<menuitem action='Stop'/>"
    "<menu action='RefreshEvery'>"
     "<menuitem action='RefreshEveryEnable'/>"
     "<separator/>"
     "<menuitem action='RefreshEvery30'/>"
     "<menuitem action='RefreshEvery60'/>"
     "<menuitem action='RefreshEvery300'/>"
     "<menuitem action='RefreshEvery900'/>"
     "<menuitem action='RefreshEvery1800'/>"
     "<menuitem action='RefreshEveryCustom'/>"
    "</menu>"
    "<separator/>"
    "<menuitem action='ZoomIn'/>"
    "<menuitem action='ZoomOut'/>"
    "<menuitem action='ZoomNormal'/>"
    "<separator/>"
    "<menuitem action='BackgroundImage'/>"
    "<menuitem action='SourceView'/>"
    "<menuitem action='Properties'/>"
    "<separator/>"
    "<menuitem action='SelectionSourceView'/>"
   "</menu>"
   "<menu action='Go'>"
    "<menuitem action='Back'/>"
    "<menuitem action='Forward'/>"
    "<menuitem action='Home'/>"
    "<menuitem action='Location'/>"
    "<menuitem action='Websearch'/>"
    // TODO: Is this a good idea?
    "<menuitem action='OpenInPageholder'/>"
    "<menu action='TabsClosed'>"
    // Closed tabs shall be prepended here
     "<separator/>"
     "<menuitem action='TabsClosedClear'/>"
    "</menu>"
    "<menuitem action='UndoTabClose'/>"
    "<separator/>"
    "<menuitem action='SelectionLinksNewTabs'/>"
    "<menuitem action='SelectionTextTabNew'/>"
    "<menuitem action='SelectionTextTabCurrent'/>"
    "<menuitem action='SelectionTextWindowNew'/>"
    "<separator/>"
    "<menuitem action='SelectionSearch'/>"
    "<menuitem action='SelectionSearchWith'/>"
   "</menu>"
   "<menu action='Bookmarks'>"
    "<menuitem action='BookmarkNew'/>"
    "<menuitem action='BookmarksManage'/>"
    "<separator/>"
    // TODO: Insert bookmarks here
   "</menu>"
   "<menu action='Tools'>"
    "<menuitem action='PanelDownloads'/>"
    "<menuitem action='PanelBookmarks'/>"
    "<menuitem action='PanelHistory'/>"
    "<menuitem action='PanelTabs'/>"
    "<menuitem action='PanelPageholder'/>"
    "<menuitem action='PanelExtensions'/>"
    "<menuitem action='PanelConsole'/>"
    "<separator/>"
    "<menuitem action='ColorPicker'/>"
    // TODO: Insert widgets and custom tools here
   "</menu>"
   "<menu action='Window'>"
    "<menuitem action='SessionLoad'/>"
    "<menuitem action='SessionSave'/>"
    "<separator/>"
    "<menuitem action='TabPrevious'/>"
    "<menuitem action='TabNext'/>"
    "<menuitem action='TabOverview'/>"
    "<separator/>"
    // All open tabs shall be appended here
   "</menu>"
   "<menu action='Help'>"
    "<menuitem action='HelpContents'/>"
    "<menuitem action='About'/>"
   "</menu>"
  "</menubar>"
  "<toolbar name='toolbar_navigation'>"
   "<toolitem action='TabNew'/>"
   "<toolitem action='Back'/>"
   "<toolitem action='Forward'/>"
   "<toolitem action='RefreshStop'/>"
   "<toolitem action='Home'/>"
   "<toolitem action='FormFill'/>"
   "<placeholder name='Location'/>"
   "<placeholder name='WebSearch'/>"
   "<placeholder name='TabTrash'/>"
  "</toolbar>"
  "<toolbar name='toolbar_panels'>"
   "<toolitem action='PanelDownloads'/>"
   "<toolitem action='PanelBookmarks'/>"
   "<toolitem action='PanelHistory'/>"
   "<toolitem action='PanelTabs'/>"
   "<toolitem action='PanelPageholder'/>"
   "<toolitem action='PanelExtensions'/>"
   "<toolitem action='PanelConsole'/>"
  "</toolbar>"
  "<popup name='popup_tabsclosed'>"
   // Closed tabs shall be prepended here
   "<separator/>"
   "<menuitem action='TabsClosedClear'/>"
  "</popup>"
  "<popup name='popup_document'>"
   "<menuitem action='Back'/>"
   "<menuitem action='Forward'/>"
   "<menuitem action='Refresh'/>"
   "<menuitem action='Stop'/>"
   "<menu action='RefreshEvery'>"
    "<menuitem action='RefreshEveryEnable'/>"
    "<separator/>"
    "<menuitem action='RefreshEvery30'/>"
    "<menuitem action='RefreshEvery60'/>"
    "<menuitem action='RefreshEvery300'/>"
    "<menuitem action='RefreshEvery900'/>"
    "<menuitem action='RefreshEvery1800'/>"
    "<menuitem action='RefreshEveryCustom'/>"
   "</menu>"
   "<separator/>"
   "<menuitem action='LinkTabNew'/>"
   "<menuitem action='LinkTabCurrent'/>"
   "<menuitem action='LinkWindowNew'/>"
   "<separator/>"
   "<menuitem action='LinkSaveAs'/>"
   "<menuitem action='LinkSaveWith'/>"
   "<menuitem action='LinkCopy'/>"
   "<menuitem action='LinkBookmarkNew'/>"
   "<separator/>"
   "<menuitem action='SelectionLinksNewTabs'/>"
   "<menuitem action='SelectionTextTabNew'/>"
   "<menuitem action='SelectionTextTabCurrent'/>"
   "<menuitem action='SelectionTextWindowNew'/>"
   "<separator/>"
   "<menuitem action='ImageViewTabNew'/>"
   "<menuitem action='ImageViewTabCurrent'/>"
   "<menuitem action='ImageSaveAs'/>"
   "<menuitem action='ImageSaveWith'/>"
   "<menuitem action='ImageCopy'/>"
   "<separator/>"
   "<menuitem action='ImageViewTabNew'/>"
   "<menuitem action='ImageViewTabCurrent'/>"
   "<separator/>"
   "<menuitem action='UndoTabClose'/>"
   "<menuitem action='SaveAs'/>"
   "<menuitem action='BookmarkNew'/>"
   "<menuitem action='Print'/>"
   "<separator/>"
   "<menuitem action='Copy'/>"
   "<menuitem action='SelectAll'/>"
   "<menuitem action='SelectionSearch'/>"
   "<menuitem action='SelectionSearchWith'/>"
   "<menuitem action='SelectionSourceView'/>"
   "<separator/>"
   "<menuitem action='BackgroundImage'/>"
   "<menuitem action='SourceView'/>"
   "<menuitem action='Properties'/>"
  "</popup>"
 "</ui>";

#endif /* !__UI_H__ */
